import { Character } from "./Character.js";
import { SlotInfo, StatusInfo } from "./definitions/adventureland.js";
import { CharacterType, CXData, DamageType, GData, MapName, NPCName, SkillName } from "./definitions/adventureland-data.js";
import { PlayerData } from "./definitions/adventureland-server.js";
import { Entity } from "./Entity.js";
export declare class Player implements PlayerData {
    protected G: GData;
    afk?: boolean | "code";
    id: string;
    ctype: CharacterType;
    abs: boolean;
    angle?: number;
    armor: number;
    apiercing: number;
    attack: number;
    avoidance: number;
    cid: number;
    crit: number;
    critdamage: number;
    damage_type: DamageType;
    evasion: number;
    frequency: number;
    going_x?: number;
    going_y?: number;
    level: number;
    move_num: number;
    moving: boolean;
    party: string;
    reflection: number;
    resistance: number;
    rpiercing: number;
    target: string;
    x: number;
    y: number;
    s: StatusInfo;
    c: any;
    cx: CXData;
    focus?: string;
    hp: number;
    max_hp: number;
    max_mp: number;
    mp: number;
    npc?: NPCName;
    owner: string;
    pdps: number;
    q: {
        compound?: {
            len: number;
            ms: number;
            num: number;
            nums: number[];
        };
        upgrade?: {
            len: number;
            ms: number;
            num: number;
        };
    };
    range: number;
    rip: boolean;
    skin: string;
    slots: SlotInfo;
    speed: number;
    stand?: boolean | "cstand" | "stand0";
    tp?: boolean;
    in: string;
    map: MapName;
    constructor(data: PlayerData, map: MapName, instance: string, g: GData);
    updateData(data: PlayerData): void;
    calculateDamageRange(defender: Character | Entity | Player, skill?: SkillName): [number, number];
    /**
     * Returns true if the player is "friendly", for example, if it's one of our characters, in our party, or in our friends list.
     * @param character Our character (e.g.: bot.character)
     */
    isFriendly(bot: Character): boolean;
    /**
     * Returns true if this player is an NPC.
     *
     * @return {*}  {boolean}
     * @memberof Player
     */
    isNPC(): boolean;
}
