import { Tools } from "./Tools.js";
export class Player {
    G;
    afk;
    id;
    ctype;
    abs;
    angle;
    armor = 0;
    apiercing = 0;
    attack;
    avoidance = 0;
    cid;
    crit = 0;
    critdamage = 0;
    damage_type;
    evasion = 0;
    frequency;
    going_x;
    going_y;
    level = 1;
    move_num;
    moving;
    party;
    reflection = 0;
    resistance = 0;
    rpiercing = 0;
    target;
    x;
    y;
    s = {};
    c = {};
    cx;
    focus;
    hp;
    max_hp;
    max_mp;
    mp;
    npc;
    owner;
    pdps;
    q = {};
    range;
    rip;
    skin;
    slots;
    speed;
    stand;
    tp;
    // Soft Properties
    in;
    map;
    constructor(data, map, instance, g) {
        this.G = g;
        // Set soft properties
        this.map = map;
        this.in = instance;
        if (!data.npc)
            this.damage_type = this.G.classes[data.ctype].damage_type;
        // Set everything else
        this.updateData(data);
    }
    updateData(data) {
        if (this.id !== undefined && this.id !== data.id)
            throw Error("The entity's ID does not match");
        // Set everything
        for (const key in data)
            this[key] = data[key];
    }
    calculateDamageRange(defender, skill = "attack") {
        // If the entity is immune, most skills won't do damage
        if (defender.immune && ["3shot", "5shot", "burst", "cburst", "supershot", "taunt"].includes(skill))
            return [0, 0];
        if (defender["1hp"] || skill == "taunt")
            return [1, 1];
        let baseDamage = this.attack;
        if (this.G.skills[skill].damage)
            baseDamage = this.G.skills[skill].damage;
        // NOTE: I asked Wizard to add something to G.conditions.cursed and .marked so we don't need these hardcoded.
        if (defender.s.cursed)
            baseDamage *= 1.2;
        if (defender.s.marked)
            baseDamage *= 1.1;
        if (this.ctype == "priest")
            baseDamage *= 0.4; // Priests only do 40% damage
        let additonalApiercing = 0;
        if (this.G.skills[skill].apiercing)
            additonalApiercing = this.G.skills[skill].apiercing;
        // NOTE: currently no skills with rpiercing
        // let additonalRpiercing = 0
        // if (this.G.skills[skill].rpiercing) additonalRpiercing = this.G.skills[skill].rpiercing
        if (this.damage_type == "physical")
            baseDamage *= Tools.damage_multiplier(defender.armor - this.apiercing - additonalApiercing);
        else if (this.damage_type == "magical")
            baseDamage *= Tools.damage_multiplier(defender.resistance - this.rpiercing /** - additionalRpiercing */);
        if (this.G.skills[skill].damage_multiplier)
            baseDamage *= this.G.skills[skill].damage_multiplier;
        let lowerLimit = baseDamage * 0.9;
        let upperLimit = baseDamage * 1.1;
        if (this.crit) {
            if (this.crit >= 100) {
                lowerLimit *= (2 + (this.critdamage / 100));
            }
            upperLimit *= (2 + (this.critdamage / 100));
        }
        // NOTE: This information is from @Wizard on Discord on May 1st, 2020
        // https://discord.com/channels/238332476743745536/243707345887166465/705722706250694737
        if (skill == "cleave") {
            lowerLimit *= 0.1;
            upperLimit *= 0.9;
        }
        return [lowerLimit, upperLimit];
    }
    /**
     * Returns true if the player is "friendly", for example, if it's one of our characters, in our party, or in our friends list.
     * @param character Our character (e.g.: bot.character)
     */
    isFriendly(bot) {
        // Check if it's us
        if (bot.id == this.id)
            return true;
        // Check if we're the owner
        if (bot.owner == this.owner)
            return true;
        // Check if we're in the same party
        if (bot.party == this.party)
            return true;
        // TODO: Check if they're in our friends list
        // They're not friendly >:(
        return false;
    }
    /**
     * Returns true if this player is an NPC.
     *
     * @return {*}  {boolean}
     * @memberof Player
     */
    isNPC() {
        return this.npc !== undefined;
    }
}
//# sourceMappingURL=Player.js.map