import { Constants } from "./Constants.js";
import { PingCompensatedCharacter } from "./PingCompensatedCharacter.js";
export class Priest extends PingCompensatedCharacter {
    ctype = "priest";
    absorbSins(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [absorbSins].");
        const absorbed = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]absorb['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`curse timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { name: "absorb", id: target });
        return absorbed;
    }
    curse(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [curse].");
        const curseStarted = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]curse['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`curse timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { name: "curse", id: target });
        return curseStarted;
    }
    darkBlessing() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [darkBlessing].");
        const darkBlessed = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]darkblessing['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`darkblessing timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { name: "darkblessing" });
        return darkBlessed;
    }
    heal(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [heal].");
        // if (!this.game.entities.has(id) && !this.game.players.has(id)) return Promise.reject(`No Entity with ID '${id}'`)
        const healStarted = new Promise((resolve, reject) => {
            const deathCheck = (data) => {
                if (data.id == id) {
                    this.socket.off("action", attackCheck);
                    this.socket.off("game_response", failCheck);
                    this.socket.off("death", deathCheck);
                    reject(`Entity ${id} not found`);
                }
            };
            const failCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "disabled") {
                        this.socket.off("action", attackCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("death", deathCheck);
                        reject(`Heal on ${id} failed (disabled).`);
                    }
                    else if (data.response == "attack_failed" && data.id == id) {
                        this.socket.off("action", attackCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("death", deathCheck);
                        reject(`Heal on ${id} failed.`);
                    }
                    else if (data.response == "too_far" && data.id == id) {
                        this.socket.off("action", attackCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("death", deathCheck);
                        reject(`${id} is too far away to heal (dist: ${data.dist}).`);
                    }
                    else if (data.response == "cooldown" && data.id == id) {
                        this.socket.off("action", attackCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("death", deathCheck);
                        reject(`Heal on ${id} failed due to cooldown (ms: ${data.ms}).`);
                    }
                }
            };
            const attackCheck = (data) => {
                if (data.attacker == this.id && data.target == id && data.type == "heal") {
                    this.socket.off("action", attackCheck);
                    this.socket.off("game_response", failCheck);
                    this.socket.off("death", deathCheck);
                    resolve(data.pid);
                }
            };
            setTimeout(() => {
                this.socket.off("action", attackCheck);
                this.socket.off("game_response", failCheck);
                this.socket.off("death", deathCheck);
                reject(`heal timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("action", attackCheck);
            this.socket.on("game_response", failCheck);
            this.socket.on("death", deathCheck);
        });
        this.socket.emit("heal", { id: id });
        return healStarted;
    }
    partyHeal() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [partyHeal].");
        const healStarted = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]partyheal['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`partyHeal timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { name: "partyheal" });
        return healStarted;
    }
    revive(target, essenceoflife = this.locateItem("essenceoflife")) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [revive].");
        if (essenceoflife === undefined)
            return Promise.reject("We don't have any essenceoflife in our inventory.");
        const revived = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]revive['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`revive timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { name: "revive", id: target, num: essenceoflife });
        return revived;
    }
}
//# sourceMappingURL=Priest.js.map