import { Constants } from "./Constants.js";
import { PingCompensatedCharacter } from "./PingCompensatedCharacter.js";
export class Rogue extends PingCompensatedCharacter {
    ctype = "rogue";
    // NOTE: UNTESTED
    // TODO: Add promises
    invis() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [invis].");
        this.socket.emit("skill", { name: "invis" });
    }
    mentalBurst(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [mentalBurst].");
        const bursted = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]mentalburst['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    this.socket.off("death", deathCheck);
                    this.socket.off("game_response", failCheck);
                    resolve();
                }
            };
            const deathCheck = (data) => {
                if (data.id == target) {
                    this.socket.off("eval", cooldownCheck);
                    this.socket.off("death", deathCheck);
                    this.socket.off("game_response", failCheck);
                    reject(`Entity ${target} not found`);
                }
            };
            const failCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "cooldown" && data.skill == "mentalburst") {
                        this.socket.off("eval", cooldownCheck);
                        this.socket.off("death", deathCheck);
                        this.socket.off("game_response", failCheck);
                        reject(`mentalBurst on ${target} failed due to cooldown (ms: ${data.ms}).`);
                    }
                    else if (data.response == "too_far" && data.place == "mentalburst") {
                        this.socket.off("eval", cooldownCheck);
                        this.socket.off("death", deathCheck);
                        this.socket.off("game_response", failCheck);
                        reject(`${target} is too far away to mentalBurst (dist: ${data.dist}).`);
                    }
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                this.socket.off("death", deathCheck);
                this.socket.off("game_response", failCheck);
                reject(`mentalBurst timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
            this.socket.on("death", deathCheck);
            this.socket.on("game_response", failCheck);
        });
        this.socket.emit("skill", {
            id: target,
            name: "mentalburst"
        });
        return bursted;
    }
    // NOTE: UNTESTED
    poisonCoat() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [poisonCoat].");
        const poisonCoated = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]pcoat['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`poisoncoat timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", {
            name: "pcoat"
        });
        return poisonCoated;
    }
    // NOTE: UNTESTED
    quickPunch(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [quickPunch].");
        const marked = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]quickpunch['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`quickPunch timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", {
            id: target,
            name: "quickpunch"
        });
        return marked;
    }
    quickStab(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [quickStab].");
        const stabbed = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]quickstab['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    this.socket.off("death", deathCheck);
                    this.socket.off("game_response", failCheck);
                    resolve();
                }
            };
            const failCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "cooldown" && data.skill == "quickstab") {
                        this.socket.off("eval", cooldownCheck);
                        this.socket.off("death", deathCheck);
                        this.socket.off("game_response", failCheck);
                        reject(`quickStab on ${target} failed due to cooldown (ms: ${data.ms}).`);
                    }
                    else if (data.response == "too_far" && data.place == "quickstab") {
                        this.socket.off("eval", cooldownCheck);
                        this.socket.off("death", deathCheck);
                        this.socket.off("game_response", failCheck);
                        reject(`${target} is too far away to quickStab (dist: ${data.dist}).`);
                    }
                }
            };
            const deathCheck = (data) => {
                if (data.id == target) {
                    this.socket.off("eval", cooldownCheck);
                    this.socket.off("death", deathCheck);
                    this.socket.off("game_response", failCheck);
                    reject(`Entity ${target} not found`);
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                this.socket.off("death", deathCheck);
                this.socket.off("game_response", failCheck);
                reject(`quickStab timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
            this.socket.on("death", deathCheck);
            this.socket.on("game_response", failCheck);
        });
        this.socket.emit("skill", {
            id: target,
            name: "quickstab"
        });
        return stabbed;
    }
    // TODO: Improve to check if we applied it on the given character
    rspeed(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [rspeed].");
        const swifted = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]rspeed['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`rspeed timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", {
            id: target,
            name: "rspeed",
        });
        return swifted;
    }
    // NOTE: UNTESTED
    shadowStrike(shadowstone = this.locateItem("shadowstone")) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [shadowStrike].");
        if (shadowstone === undefined)
            return Promise.reject("We need a shadowstone in order to shadowstrike.");
        const shadowStriked = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]shadowstrike['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("eval", cooldownCheck);
                reject(`shadowstrike timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", {
            name: "shadowstrike",
            num: shadowstone
        });
        return shadowStriked;
    }
    // NOTE: Untested
    // TODO: Add promises
    stopInvis() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [stopInvis].");
        this.socket.emit("stop", { action: "invis" });
    }
}
//# sourceMappingURL=Rogue.js.map