/**
 * This file aims to contain definitions for socket related things that the main
 * game uses to interact with the server.
 */
import { StatusInfo, SlotInfo, ServerRegion, ServerIdentifier, BankInfo, TradeSlotType } from "./adventureland.js";
import { AchievementName, AnimationName, Attribute, CharacterType, CXData, EmotionName, GDropItem, ItemName, MapName, MonsterName, NPCName, ProjectileName, SkillName, TitleName } from "./adventureland-data.js";
export declare type AchievementProgressData = AchievementProgressDataFirehazard | {
    name: string;
};
export declare type AchievementProgressDataFirehazard = {
    name: "firehazard";
    count: number;
    needed: number;
};
export declare type ActionData = {
    attacker: string;
    damage?: number;
    heal?: number;
    eta: number;
    m: number;
    pid: string;
    projectile: ProjectileName;
    source: SkillName;
    target: string;
    type: SkillName;
    x: number;
    y: number;
};
export declare type AuthData = {
    auth: string;
    /** NOTE: This is not the name of the player. It's a long number, encoded as a string. */
    user: string;
    /** NOTE: This is not the name of the character. It's a long number, encoded as a string. */
    character: string;
    passphrase: string;
    height: number;
    width: number;
    scale: number;
    no_html: "" | "1";
    no_graphics: "" | "True";
    code_slot?: number;
};
export declare type CharacterData = PlayerData & {
    hp: number;
    max_hp: number;
    mp: number;
    max_mp: number;
    attack: number;
    fear: number;
    courage: number;
    mcourage: number;
    pcourage: number;
    frequency: number;
    speed: number;
    range: number;
    armor: number;
    resistance: number;
    level: number;
    rip: boolean;
    afk: boolean | string | "afk";
    s: StatusInfo;
    c: any;
    q: {
        compound?: {
            len: number;
            ms: number;
            num: number;
            nums: number[];
        };
        upgrade?: {
            len: number;
            ms: number;
            num: number;
        };
        exchange?: {
            len: number;
            ms: number;
        };
    };
    /** TODO: What is this? */
    abs?: boolean;
    age: number;
    angle?: number;
    blast: number;
    pdps: number;
    id: string;
    name?: string;
    x: number;
    y: number;
    going_x?: number;
    going_y?: number;
    moving?: boolean;
    stand?: boolean | "cstand" | "stand0";
    skin: string;
    slots: SlotInfo;
    ctype: CharacterType;
    owner: string;
    party?: string;
    explosion: number;
    firesistance: number;
    fzresistance: number;
    mp_reduction: number;
    pnresistance: number;
    stun: number;
    int: number;
    str: number;
    dex: number;
    vit: number;
    for: number;
    mp_cost: number;
    max_xp: number;
    goldm: number;
    xpm: number;
    xp: number;
    luckm: number;
    map: MapName;
    in: string;
    /** The size of the character's inventory */
    isize: number;
    /** The number of empty inventory slots */
    esize: number;
    gold: number;
    cash: number;
    /** This number is the number of monsters currently targeting you */
    targets: number;
    target?: string;
    m: number;
    evasion: number;
    miss: number;
    move_num?: number;
    reflection: number;
    lifesteal: number;
    manasteal: number;
    rpiercing: number;
    apiercing: number;
    crit: number;
    critdamage: number;
    dreturn: number;
    emx?: {
        [T in EmotionName]?: number;
    };
    tax: number;
    xrange: number;
    items: ItemData[];
    cc: number;
    cid: number;
    controller?: string;
    cx: CXData;
    ipass?: string;
    friends?: string[];
    acx?: any;
    xcx?: string[];
    /** Extra events (e.g. ["game_response", {response: "upgrade_success", level: 4, num: 8}]) */
    hitchhikers?: [string, any][];
    /** Holds bank information when the character is inside the bank */
    user?: BankInfo;
    /** (GUI Related) Set if you move inventory items. Flag for reopening player's inventory. */
    reopen?: boolean;
};
export declare type CharacterListData = {
    map: MapName;
    in: string;
    name: string;
    level: number;
    skin: string;
    cx: CXData;
    online: number;
    y: number;
    x: number;
    type: CharacterType;
    id: string;
    secret?: string;
    server?: string;
};
export declare type ChestData = {
    chest: string | "chest3" | "chest4" | "chest6";
    id: string;
    items: number;
    map: MapName;
    party: string;
    x: number;
    y: number;
};
export declare type ChestOpenedData = {
    id: string;
    gold: number;
    goldm: number;
    items: {
        name: string;
        q?: number;
        level?: number;
        looter: string;
    }[];
    opener: string;
    party: boolean;
} | {
    id: string;
    gone: true;
};
export declare type CMData = {
    /** The name of the player that sent the CM */
    name: string;
    /** The message that they sent */
    message: string;
};
export declare type DeathData = {
    id: string;
    place?: string | "attack";
};
export declare type DisappearData = 
/** Character used 'blink' */
{
    /** Blink animation will be used */
    effect: "blink";
    /** Character name */
    id: string;
    reason: "transport";
    /** [x, y, orientation (up/down/left/right)] */
    s?: [number, number, number];
    to?: MapName;
} | 
/** Character (rogue) went invisible */
{
    /** Character name */
    id: string;
    invis: true;
    reason: "invis";
} | 
/** Character disconnected */
{
    /** Character name */
    id: string;
    reason: "disconnect";
} | 
/** Character used 'magiport' */
{
    /** Magiport animation will be used */
    effect: "magiport";
    /** Character name */
    id: string;
    reason: "transport";
    /** [x, y] */
    s?: [number, number];
    to?: MapName;
} | 
/** Character went through a door */
{
    /** No animation will be used */
    effect?: undefined;
    /** Character name */
    id: string;
    reason: "transport";
    /** The map spawn where the character will appear */
    s?: number;
    /** The map the character went to */
    to?: MapName | string;
} | 
/** Character used a 'town' teleport */
{
    /** Town teleport animation will be used */
    effect: 1;
    /** Character name */
    id: string;
    reason: "transport";
    s?: number;
    to?: MapName;
};
export declare type DisappearingTextData = {
    message: string;
    x: number;
    y: number;
    id: string;
    args: any;
};
export declare type DisconnectReasonData = "limitdc";
export declare type EmotionData = {
    /** emotion name */
    name: EmotionName;
    /** character name that did the emotion */
    player: string;
};
export declare type EntitiesData = {
    type: "all" | "xy";
    in: string;
    map: MapName;
    monsters: MonsterData[];
    players: PlayerData[];
};
export declare type FriendData = {
    event: "lost";
    friends: string[];
    /** The name of the player that you are no longer friends with */
    name: string;
} | {
    event: "new";
    /** The name of the player that you are now friends with */
    friends: string[];
    name: string;
} | {
    event: "request";
    /** The name of the player who sent you a friend request */
    name: string;
} | {
    event: "update";
    friends: string[];
};
export declare type MonsterData = {
    id: string;
    type: MonsterName;
    angle?: number;
    move_num?: number;
    moving?: boolean;
    x: number;
    y: number;
    armor?: number;
    attack?: number;
    cid: number;
    frequency?: number;
    going_x: number;
    going_y: number;
    mp: number;
    speed?: number;
    resistance?: number;
    s: StatusInfo;
    /** The ID of the target */
    target?: string;
    abs?: false;
    hp?: number;
    level?: number;
    max_hp?: number;
    xp?: number;
};
export declare type EvalData = {
    code: string;
};
export declare type LoginData = LoginDataItem[];
export declare type LoginDataItem = {
    code: string;
    type: string;
} | {
    message: string;
    type: string;
} | {
    html: string;
};
export declare type GameLogData = GameLogDataString;
export declare type GameLogDataString = string | "Already partying" | "Can't respawn yet." | "Invitation expired";
export declare type GameResponseData = GameResponseDataObject | GameResponseDataString;
export declare type GameResponseDataObject = {
    response: "attack_failed";
    place: "attack";
    id: string;
} | 
/** When you try to enter the bank, but another one of your characters is already inside. */
{
    response: "bank_opx";
    /** The character that is already inside */
    name: string;
    reason: "mounted";
} | {
    response: "bank_restrictions";
    place: string | "compound";
} | {
    response: "buy_success";
    cost: number;
    num: number;
    name: ItemName;
    q: number;
} | {
    response: "cooldown";
    skill?: SkillName;
    place: SkillName;
    id: string;
    ms: number;
} | {
    response: "craft";
    name: ItemName;
} | {
    response: "defeated_by_a_monster";
    xp: number;
    monster: MonsterName;
} | {
    response: "disabled";
    place: "attack";
} | 
/** Called when a condition expires */
{
    response: "ex_condition";
    name: SkillName;
} | {
    response: "get_closer";
    place: "upgrade";
} | {
    response: "gold_sent";
    name: string;
    gold: number;
} | {
    response: "item_locked";
    place: "upgrade";
} | {
    response: "item_sent";
    name: string;
    item: ItemName;
    q: number;
} | {
    response: "magiport_failed";
    id: string;
} | {
    response: "magiport_sent";
    id: string;
} | {
    response: "mail_item_taken";
} | {
    response: "no_mp";
    place: SkillName;
} | {
    response: "no_target";
} | {
    response: "seashell_success";
    suffix: string | "";
} | {
    response: "skill_fail";
    name: SkillName;
} | {
    response: "skill_success";
    name: SkillName;
} | {
    response: "target_lock";
    monster: MonsterName;
} | {
    response: "too_far";
    place: SkillName;
    id: string;
    dist: number;
} | 
/** When you try to unfriend, but you have a character in the bank */
{
    response: "unfriend_failed";
    reason: "bank" | "coms failure" | "nouser";
} | {
    response: "gold_received" | "item_placeholder" | "item_received";
    gold: number;
    name: string;
};
export declare type GameResponseDataString = "bank_restrictions"
/** When you attempt to blink to a spot you can't reach */
 | "blink_failed" | "buy_cant_npc" | "buy_cant_space" | "buy_cost"
/** When you're too far from Ponty and try to view Ponty's items */
 | "buy_get_closer"
/** When you try to use `transport` on a door that needs `enter` */
 | "cant_enter"
/** When attempting to leave a map you can't use the leave command on */
 | "cant_escape"
/** ??? Maybe if we attempt to compound something with an inventory position that is empty ??? */
 | "compound_no_item"
/** When we try to dash too far */
 | "dash_failed"
/** Too far away from monster hunt npc */
 | "ecu_get_closer"
/** When you try to do an emotion but it's not a valid emotion name, or you don't have that emotion */
 | "emotion_cant"
/** When you try to do an emotion but it's rejected because it's on cooldown */
 | "emotion_cooldown"
/** We are already exchanging something */
 | "exchange_existing"
/** The given item requires multiple to exchange */
 | "exchange_notenough"
/** When you send a friend request for someone you're already friends with */
 | "friend_already"
/** When you try to accept a friend request, but you took too long (or they never sent one in the first place) */
 | "friend_expired"
/** When you send a friend request but they aren't on the server to accept the request */
 | "friend_rleft"
/** When you send a friend request for a valid player name */
 | "friend_rsent"
/** When a merchant tries to start a monster hunt */
 | "monsterhunt_merchant" | "monsterhunt_started"
/** When you try to use a skill, but you're not a high enough level */
 | "no_level"
/** When you attack or use a skill with "id" set to "null" */
 | "no_target"
/** After you use a skill, when the server is done with everything, it will send this in response */
 | "resolve_skill"
/** When you try to send an item to another character, but they don't have room for it in their inventory */
 | "send_no_space" | "skill_cant_incapacitated"
/** When you try to use a skill, but you don't have the right weapon type equipped for that skill */
 | "skill_cant_wtype" | "skill_too_far"
/** When you try to list an item for sale in a slot that already has something listed */
 | "slot_occupied"
/** When you try to sell an item to another merchant, but there's no space on that merchant */
 | "trade_bspace" | "trade_get_closer"
/** When you try to enter a dungeon, but you don't have a key */
 | "transport_cant_item"
/** When you try to go through a door you haven't unlocked yet (e.g. lower bank) */
 | "transport_cant_locked"
/** When you're too far away from a door */
 | "transport_cant_reach"
/** Failed upgrading (to chance) */
 | "upgrade_fail"
/** We are already upgrading something */
 | "upgrade_in_progress"
/** We are trying to use a scroll to upgrade something that is a higher grade than the scroll can upgrade */
 | "upgrade_incompatible_scroll"
/** Successfully upgraded */
 | "upgrade_success";
export declare type HitData = {
    anim?: AnimationName | "miss" | "reflect";
    /** If this is set, we avoided the projectile (by running?) */
    avoid?: boolean;
    damage?: number;
    evade?: boolean;
    heal?: number;
    hid?: string;
    id?: string;
    /** Did the entity die from this hit? */
    kill?: boolean;
    lifesteal?: number;
    manasteal?: number;
    miss?: boolean;
    pid?: string;
    projectile?: string;
    reflect?: number;
    /** If set, this was a sneak attack by a rogue */
    sneak?: boolean;
    source?: SkillName | "burn";
    /** If this is set, these IDs are too close to each other and are receiving additional damage on each hit */
    stacked?: string[];
    /** If set, the character is stunned with this attack */
    stun?: boolean;
};
/** Used for the 'invite' socket message */
export declare type InviteData = {
    /** The name of the character who invited */
    name: string;
};
export declare type ItemData = {
    /** Achievement progress */
    acc?: number;
    /** Achievement name to which you are progressing to */
    ach?: AchievementName;
    /** For cxjars, it shows the cosmetic contained inside. */
    data?: string;
    /** TODO: What is this? Seen on an elixir. TODO: Check if this is only applicable to the elixir slot */
    ex?: boolean;
    /** If this is set, once the date hits, the item will disappear */
    expires?: string;
    /** TODO: What is this? Seen on a level 1 xpbooster. */
    extra?: number;
    /** TODO: Confirm. The item was a giveaway, and this was the character that gave it away. */
    gf?: string;
    /** If set, the item was given to the player (most likely when they created the character), and it is only worth 1 gold if you sell it to an NPC. */
    gift?: number;
    /** The higher the number, the more likely it is to succeed if you compound or upgrade the item */
    grace?: number;
    /** If set, the item is locked. 's' == 'sealed', 'u' == 'unlocking', "l" == 'locked'. */
    l?: "l" | "s" | "x";
    /** The item level */
    level?: number;
    /** The item name */
    name: ItemName;
    /** The title applied to the item */
    p?: TitleName;
    /** TODO: What is this? A list of available titles for the item?  */
    ps?: TitleName[];
    /** The quantity of the item for stackable items */
    q?: number;
    /** If the item has a scroll applied to it, the scroll provides this attribute */
    stat_type?: Attribute;
    /** If set, the item is PVP marked until the given date. If you die to another player, there is a chance to lose this item to them. */
    v?: string;
};
/**
 * This is for items that are in trade slots
 */
export declare type ItemDataTrade = ItemData & {
    /** If set, this item is not for sale. The player wants to buy this item. */
    b?: boolean;
    /** Number of minutes remaining for giveaway items */
    giveaway?: number;
    /** List of character IDs that are in the giveaway */
    list?: string[];
    price: number;
    rid: string;
};
export declare type LimitDCReportData = {
    /** How many of each call you made */
    mcalls: {
        auth?: number;
        code?: number;
        loaded?: number;
        move?: number;
        "o:home"?: number;
        ping_trig?: number;
        players?: number;
        property?: number;
        render?: number;
        send_updates?: number;
    };
    /** Call cost limit. It's lower for comm sockets. */
    climit?: number;
    /** Total number of socket messages sent */
    total?: number;
};
export declare type LoadedData = {
    /** The height of the monitor's resolution */
    height: number;
    /** The width of the monitor's resolution */
    width: number;
    scale: number;
    success: number | 1;
};
export declare type MailData = {
    /** We can use this cursor to retrieve more mail */
    cursor: string;
    /** If false, we are on the first page of mail. If true, we are scrolling through mail */
    cursored: boolean;
    mail: MailMessageData[];
    /** If true, there is more mail which you can retrieve by using the cursor */
    more: boolean;
    type: "mail";
};
export declare type MailMessageData = {
    /** The mail's ID */
    id: string;
    /** From */
    fro: string;
    /** To */
    to: string;
    /** The time this message was sent */
    sent: string;
    /** If true, we have taken the item attached to this mail. */
    message: string;
    /** A string containing a JSON object that represents the item that was sent in the mail */
    item?: string;
    taken?: boolean;
};
export declare type PullMerchantsData = {
    type: "merchants";
    chars: PullMerchantsCharData[];
};
export declare type PullMerchantsCharData = {
    cx: CXData;
    map: MapName;
    x: number;
    y: number;
    afk: boolean | "code";
    /** Follows the following format: `${ServerRegion} ${ServerIdentifier}` */
    server: string;
    stand: "cstand" | "stand0" | "stand1";
    /** Merchant name */
    name: string;
    /** Merchant level */
    level: number;
    /** Items for sale */
    slots: {
        [T in TradeSlotType]?: ItemDataTrade;
    };
    skin: string;
};
export declare type NewMapData = {
    direction: number;
    effect: number | "blink" | "magiport";
    entities: EntitiesData;
    in: string;
    info: any;
    m: number;
    name: MapName;
    x: number;
    y: number;
};
export declare type NotThereData = {
    place: "attack";
};
export declare type OnlineData = {
    name: string;
    server: string;
};
export declare type PartyData = {
    list: string[];
    message?: string;
    party: {
        [T in string]: {
            gold: number;
            in: string;
            l: number;
            level: number;
            luck: number;
            map: MapName;
            share: number;
            pdps: number;
            skin: string;
            type: CharacterType;
            x: number;
            xp: number;
            y: number;
            cx?: CXData;
        };
    };
};
export declare type PlayerData = {
    id: string;
    ctype: CharacterType | NPCName;
    abs?: boolean;
    angle?: number;
    going_x?: number;
    going_y?: number;
    /** The 'pvp' NPC has an extra  */
    allow?: boolean;
    afk?: boolean | "code";
    age?: number;
    armor: number;
    attack?: number;
    c: any;
    cid: number;
    code?: boolean | string;
    controller?: string;
    cx: CXData;
    focus?: string;
    frequency?: number;
    x: number;
    y: number;
    hp: number;
    level: number;
    max_hp: number;
    max_mp?: number;
    move_num?: number;
    moving?: boolean;
    mp?: number;
    npc?: NPCName;
    owner: string;
    party?: string;
    pdps?: number;
    q: {
        compound?: {
            len: number;
            ms: number;
            num: number;
            nums: number[];
        };
        exchange?: {
            ms: number;
            len: number;
            name: ItemName;
            id: ItemName;
            q: number;
            num: number;
        };
        upgrade?: {
            len: number;
            ms: number;
            num: number;
        };
    };
    range?: number;
    resistance?: number;
    rip?: boolean | number;
    s: StatusInfo;
    skin: string;
    slots?: SlotInfo;
    speed: number;
    stand?: boolean | "cstand" | "stand0";
    target?: string;
    tp?: boolean;
    xp?: number;
};
/**
 * This is what is returned when you ask for 'players'.
 * It's all the players that are online on the server you're on.
 */
export declare type PlayersData = {
    /** Are they AFK? */
    afk: number;
    /** How many days have passed since this character was created */
    age: number;
    /** What level are they */
    level: number;
    /** Character name */
    name: string;
    /** What map are they on */
    map: MapName;
    /** What party are they in? */
    party: string;
    /** What type of character are they */
    type: CharacterType;
}[];
export declare type QData = {
    num: number;
    p: {
        chance: number;
        level: number;
        name: ItemName;
        nums: number[];
        scroll: ItemName;
        success?: boolean;
    };
    q: {
        compound?: {
            len: number;
            ms: number;
            num: number;
            nums: number[];
        };
        upgrade?: {
            len: number;
            ms: number;
            num: number;
        };
    };
};
export declare type ServerData = {
    addr: string;
    port: number;
    region: ServerRegion;
    name: ServerIdentifier;
    players: number;
    key: string;
};
export declare type ServerInfoData = {
    [T in MonsterName]?: ServerInfoDataLive | ServerInfoDataNotLive;
} & {
    egghunt?: boolean;
    halloween?: boolean;
    holidayseason?: boolean;
    lunarnewyear?: boolean;
    valentines?: boolean;
};
export declare type ServerInfoDataLive = {
    hp: number;
    live: true;
    map: MapName;
    max_hp: number;
    target?: string;
    /** NOTE: Some event monsters don't have x and y (e.g.: Slenderman) */
    x?: number;
    /** NOTE: Some event monsters don't have x and y (e.g.: Slenderman) */
    y?: number;
};
export declare type ServerInfoDataNotLive = {
    live: false;
    /** When the monster will spawn next */
    spawn: string;
};
export declare type StartData = CharacterData & {
    info: any;
    code_slot: number;
    code_version: number;
    base_gold: {
        [T in MonsterName]?: {
            [T in string]?: number;
        };
    };
    s_info: ServerInfoData;
    entities: EntitiesData;
};
export declare type TrackerData = {
    /** How many points you have for each monster for the given character */
    monsters: {
        [T in MonsterName]?: number;
    };
    /** How many points behind this character is from your character that has the most points for this monster */
    monsters_diff: {
        [T in MonsterName]?: number;
    };
    /** How many exchanges this character has done for the various given items */
    exchanges: {};
    /** Contains drop information */
    maps: {
        [T in MapName | "global_static" | "global"]?: GDropItem[];
    };
    /** For the "open" items in maps, this table has a list of the drops that could occur */
    tables: {
        [T in ItemName | string]?: GDropItem[];
    };
    /** Contains information about your characters with the max points */
    max: {
        monsters: {
            [T in MonsterName]?: [number, string];
        };
    };
    drops: {
        [T in MonsterName]?: GDropItem[];
    };
    global: [number, "open", string][] | [number, ItemName][];
    global_static: [number, "open", string][] | [number, ItemName][];
};
export declare type UIData = {
    type: "-$";
    id: string | "basics" | "scrolls";
    /** The character who sold the item */
    name: string;
    item: {
        name: ItemName;
        q: number;
    };
    /** The slot number where the item was being stored in inventory. */
    num: string;
} | {
    type: "+$";
    id: string | "basics" | "scrolls";
    /** The character who bought the item */
    name: string;
    item: {
        name: ItemName;
        q: number;
    };
} | {
    type: "+$$";
    seller: string;
    buyer: string;
    item: ItemData & {
        price: number;
    };
    slot: TradeSlotType;
    num: number;
    snum: number;
} | {
    type: "fishing_fail";
    name: string;
} | {
    type: "fishing_none";
} | {
    type: "fishing_start";
    name: string;
    direction: number;
} | {
    type: "massproduction";
    name: string;
} | {
    type: "mining_fail";
    name: string;
} | {
    type: "mining_none";
} | {
    type: "mining_start";
    name: string;
    direction: number;
} | {
    type: "mluck";
    from: string;
    to: string;
} | {
    type: "scare";
    name: string;
    ids: string[];
};
export declare type UpgradeData = {
    type: string | "compound" | "exchange" | "upgrade";
    /** 0 = fail, 1 = success */
    success: 0 | 1;
};
export declare type WelcomeData = {
    /** The character gets returned if you open a socket with a secret (i.e. /comm and click on one of your characters) */
    character?: CharacterData;
    region: ServerRegion;
    in: string;
    map: MapName;
    gameplay: string | "normal";
    info: any;
    name: ServerIdentifier;
    pvp: boolean;
    x: number;
    y: number;
};
